// file      : xsd/cxx-cli/elements.cxx-cli
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx-cli/elements.hxx>

#include <backend-elements/regex.hxx>

#include <sstream>
#include <iostream>

using std::wcerr;
using std::endl;

namespace CXX_CLI
{
  // Context
  //

  String Context::
  ns_name (SemanticGraph::Namespace& ns)
  {
    using BackendElements::Regex::perl_s;

    using SemanticGraph::Schema;
    using SemanticGraph::Includes;
    using SemanticGraph::Imports;
    using SemanticGraph::Implies;
    using SemanticGraph::Sources;

    SemanticGraph::Path file;

    Schema& schema (dynamic_cast<Schema&> (ns.scope ()));

    if (schema.contained ())
    {
      SemanticGraph::Contains& c (*schema.contained_begin ());

      if (Includes* i = dynamic_cast<Includes*> (&c)) file = i->file (); else
      if (Imports* i = dynamic_cast<Imports*> (&c)) file = i->file (); else
      if (Implies* i = dynamic_cast<Implies*> (&c)) file = i->file (); else
      if (Sources* s = dynamic_cast<Sources*> (&c)) file = s->file ();
    }

    String tmp, pair (file.empty () ? "" : file.string ());

    pair += L' ' + ns.name ();

    if (trace_namespace_regex)
      wcerr << "pat: '" << pair << "'" << endl;

    bool found (false);

    for (NamespaceMapping::ConstReverseIterator e (ns_mapping.rbegin ());
         e != ns_mapping.rend (); ++e)
    {
      tmp = perl_s (pair, *e);
      tmp = perl_s (tmp, L"#/#::#"); // replace `/' with `::'

      // Check the result.
      //
      if (perl_s (tmp, L"/^([a-zA-Z_]\\w*)(::[a-zA-Z_]\\w*)*$//").empty ())
        found = true;

      if (trace_namespace_regex)
        wcerr << "try: '" << *e << "' : '" << tmp << "' : "
              << (found ? '+' : '-') << endl;

      if (found)
        break;
    }

    if (!found)
    {
      // Check if the name is valid by itself.
      //

      if (ns.name ().empty ())
      {
        // Empty name denotes a no-namespace case.
        //
        tmp = ns.name ();
      }
      else
      {
        tmp = perl_s (ns.name (), L"#/#::#"); // replace `/' with `::'

        if (!perl_s (tmp, L"/^([a-zA-Z_]\\w*)(::[a-zA-Z_]\\w*)*$//").empty ())
        {
          throw NoNamespaceMapping (
            ns.file (), ns.line (), ns.column (), ns.name ());
        }
      }
    }

    // Parse resulting namespace string and id() each name.
    //
    String r;
    String::size_type b (0), e;

    do
    {
      e = tmp.find (L"::", b);

      String  name (tmp, b, e == tmp.npos ? e : e - b);

      if (!name.empty ())
        r += L"::" + escape (name);

      b = e;

      if (b == tmp.npos)
        break;

      b += 2;

    } while (true);

    return r;
  }

  SemanticGraph::Namespace& Context::
  namespace_ (SemanticGraph::Nameable& n)
  {
    // The basic idea goes like this: go up Names edges until you
    // reach Namespace. There are, however, anonymous types which
    // need special handling. In the case of an anonymous type we
    // will go up Belongs edge.
    //

    if (n.named ())
    {
      SemanticGraph::Scope& s (n.scope ());

      SemanticGraph::Namespace* ns (
        dynamic_cast<SemanticGraph::Namespace*> (&n));

      return ns ? *ns : namespace_ (s);
    }
    else
    {
      SemanticGraph::Type& t (dynamic_cast<SemanticGraph::Type&> (n));

      SemanticGraph::Belongs& b (*t.classifies_begin ());

      return namespace_ (b.instance ());
    }
  }

  String Context::
  xs_ns_name (SemanticGraph::Nameable& n)
  {
    return namespace_ (n).name ();
  }

  String Context::
  fq_name (SemanticGraph::Nameable& n)
  {
    using namespace SemanticGraph;

    String r;

    if (dynamic_cast<Schema*> (&n))
    {
      return L""; // Map to global namespace.
    }
    else if (SemanticGraph::Namespace* ns =
             dynamic_cast<SemanticGraph::Namespace*> (&n))
    {
      r = ns_name (*ns);
    }
    else
    {
      r = fq_name (n.scope ());
      r += L"::";
      r += escape (n.name ());

      // I am using escape() here while theoretically I should be
      // using id() and recreate the while enter_scope/leave_scope
      // thing. I don't do it right now since we don't "rename" types
      // and I don't use fq_name enywhere else. But if things change
      // then I will have to take care this.
      //
    }

    return r;
  }

  namespace
  {
    wchar_t* keywords[] = {
      L"and",
      L"asm",
      L"auto",
      L"bitand",
      L"bitor",
      L"bool",
      L"break",
      L"case",
      L"catch",
      L"char",
      L"class",
      L"compl",
      L"const",
      L"const_cast",
      L"continue",
      L"default",
      L"delete",
      L"do",
      L"double",
      L"dynamic_cast",
      L"else",
      L"end_eq",
      L"enum",
      L"explicit",
      L"export",
      L"extern",
      L"false",
      L"float",
      L"for",
      L"friend",
      L"goto",
      L"if",
      L"inline",
      L"int",
      L"long",
      L"mutable",
      L"namespace",
      L"new",
      L"not",
      L"not_eq",
      L"operator",
      L"or",
      L"or_eq",
      L"private",
      L"protected",
      L"public",
      L"register",
      L"reinterpret_cast",
      L"return",
      L"short",
      L"signed",
      L"sizeof",
      L"static",
      L"static_cast",
      L"struct",
      L"switch",
      L"template",
      L"this",
      L"throw",
      L"true",
      L"try",
      L"typedef",
      L"typeid",
      L"typename",
      L"union",
      L"unsigned",
      L"using",
      L"virtual",
      L"void",
      L"volatile",
      L"wchar_t",
      L"while",
      L"xor",
      L"xor_eq"
    };
  }

  String Context::
  escape (String const& name) const
  {
    Size const size (sizeof (keywords) / sizeof (WideChar*));

    if (std::binary_search (keywords, keywords + size, name))
      return name + L'_';

    String r (name);

    if (!std::isalpha (r[0], locale) && r[0] != L'_')
      r = L"cxx_" + r;

    for (String::Iterator i (r.begin () + 1), e (r.end ()); i != e; ++i)
      if (!std::isalnum (*i, locale) && *i != L'_')
        *i = L'_';

    return r;
  }

  String Context::
  id (Char const* p, String const& n, Char const* s) const
  {
    String prefix (p), suffix (s);

    String core (n);
    String escaped (escape (prefix + escape (core) + suffix));

    if (!scope_stack.empty ())
    {
      Scope const& scope (scope_stack.back ());

      UnsignedLong i (0);  // 0 means this the orginal name.
      Boolean conflict (false);

      do
      {
        conflict = false;

        // If this is a class scope, make sure the name doesn't conflict
        // with the class name.
        //
        if (scope.class_scope)
        {
          String n (escape (prefix + scope.name + suffix));

          if (escaped == n)
            conflict = true;
        }


        // If this is a class scope, make sure the name doesn't conflict
        // with other member names. We need to do this since the name
        // can be escaped due to a conflict above.
        //
        if (!conflict && scope.class_scope && scope.scope != 0)
        {

          // This is a bit tricky. If this the original name (i.e., i == 0)
          // then one Nameable in the scope indicates that there is no
          // conflict. If i > 0 then any Nameable in the scope indicate
          // that there is a conflic.
          //
          SemanticGraph::Scope::NamesIteratorPair r (
            scope.scope->find (core));

          // List doesn't have operator-.
          //
          UnsignedLong d (0);

          for (; r.first != r.second; ++r.first)
            ++d;

          if ((i == 0 && d > 1) || (i > 0 && d > 0))
          {
            conflict = true;
          }
        }

        if (conflict)
        {
          std::wostringstream os;
          os << ++i;
          core = n + os.str ();
          escaped = escape (prefix + escape (core) + suffix);
        }
      }
      while (conflict);
    }

    return escaped;
  }

  // Namespace
  //

  Void Namespace::
  pre (Type& n)
  {
    String ns (ctx_.ns_name (n));

    String::size_type b (0), e;

    if (set_scope_)
      ctx_.enter_scope (L"");

    do
    {
      e = ns.find (L"::", b);

      String name (ns, b, e == ns.npos ? e : e - b);

      if (!name.empty ())
      {
        ctx_.os << "namespace " << name << "{";

        if (set_scope_)
          ctx_.enter_scope (name);
      }


      b = e;

      if (b == ns.npos)
        break;

      b += 2;

    } while (true);
  }

  Void Namespace::
  post (Type& n)
  {
    String ns (ctx_.ns_name (n));

    String::size_type b (0), e;

    do
    {
      e = ns.find (L"::", b);

      String name (ns, b, e == ns.npos ? e : e - b);

      if (!name.empty ())
      {
        ctx_.os << "}";

        if (set_scope_)
          ctx_.leave_scope ();
      }


      b = e;

      if (b == ns.npos)
        break;

      b += 2;

    }
    while (true);

    if (set_scope_)
      ctx_.leave_scope ();
  }
}
