// file      : xsd/cxx-cli/parser/elements.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#ifndef CXX_CLI_PARSER_ELEMENTS_HXX
#define CXX_CLI_PARSER_ELEMENTS_HXX

#include <sstream>

#include <cxx-cli/elements.hxx>

#include <cxx-cli/parser/cli.hxx>

namespace CXX_CLI
{
  namespace Parser
  {
    //
    //
    class Context: public CXX_CLI::Context
    {
    public:
      Context (std::wostream& o, CLI::Options const& ops)
          : CXX_CLI::Context (o,
                          ops.value<CLI::char_type> (),
                          ops.value<CLI::include_with_brackets> (),
                          ops.value<CLI::include_prefix> (),
                          ops.value<CLI::export_symbol> (),
                          ops.value<CLI::namespace_regex> (),
                          ops.value<CLI::namespace_regex_trace> (),
                          ops.value<CLI::generate_inline> ()),
            parser_type (parser_type_),
            xml_element_type (xml_element_type_),
            xml_attribute_type (xml_attribute_type_),
            parser_type_ (L"::xsd::cxx_cli::parser< " + char_type + L" >"),
            xml_element_type_ (L"::xsd::cxx_cli::xml::element< " + char_type + L" >"),
            xml_attribute_type_ (L"::xsd::cxx_cli::xml::attribute< " + char_type + L" >")
      {
      }

    protected:
      Context (Context& c)
          : CXX_CLI::Context (c),
            parser_type (c.parser_type),
            xml_element_type (c.xml_element_type),
            xml_attribute_type (c.xml_attribute_type)
      {
      }

      Context (Context& c, std::wostream& o)
          : CXX_CLI::Context (c, o),
            parser_type (c.parser_type),
            xml_element_type (c.xml_element_type),
            xml_attribute_type (c.xml_attribute_type)
      {
      }

    public:
      String& parser_type;
      String& xml_element_type;
      String& xml_attribute_type;

    private:
      String parser_type_;
      String xml_element_type_;
      String xml_attribute_type_;
    };


    //
    //
    struct TemplateParamMemberDecl : Traversal::Member,
                                     protected virtual Context
    {
      TemplateParamMemberDecl (Context& c)
          : Context (c)
      {
      }

      virtual Void
      traverse (Type& m)
      {
        os << ", typename " << id ("_xsd_", m.name ());
      }
    };


    //
    //
    struct TemplateParamBaseDecl : Traversal::Complex,
                                   protected virtual Context
    {
      TemplateParamBaseDecl (Context& c)
          : Context (c), member_ (c)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> member_;
      }

      virtual Void
      traverse (Type& c)
      {
        os << ", typename " << id ("_xsd_", c.name ());

        enter_class_scope (id (c.name ()), c);
        Traversal::Complex::traverse (c);
        leave_scope ();
      }

    private:
      Traversal::Inherits inherits_;

      TemplateParamMemberDecl member_;
      Traversal::Names names_;
    };


    //
    //
    struct TemplateParamMemberUse : Traversal::Member,
                                    protected virtual Context
    {
      TemplateParamMemberUse (Context& c)
          : Context (c)
      {
      }

      virtual Void
      traverse (Type& m)
      {
        os << ", " << id ("_xsd_", m.name ());
      }
    };


    //
    //
    struct TemplateParamBaseUse : Traversal::Complex,
                                  protected virtual Context
    {
      TemplateParamBaseUse (Context& c)
          : Context (c), member_ (c)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> member_;
      }

      virtual Void
      traverse (Type& c)
      {
        os << ", " << id ("_xsd_", c.name ());

        enter_class_scope (id (c.name ()), c);
        Traversal::Complex::traverse (c);
        leave_scope ();
      }

    private:
      Traversal::Inherits inherits_;

      TemplateParamMemberUse member_;
      Traversal::Names names_;
    };


    //
    //
    struct TemplateParamDeclList: Traversal::Complex,
                                  protected virtual Context
    {
      TemplateParamDeclList (Context& c)
          : Context (c),
            template_param_member_ (c),
            template_param_base_ (c)
      {
        template_param_names_ >> template_param_member_;
        template_param_inherits_ >> template_param_base_;
      }

      virtual Void
      traverse (Type& c)
      {
        os << "typename " << id ("_xsd_", c.name ());

        enter_class_scope (id (c.name ()), c);

        inherits (c, template_param_inherits_);
        names (c, template_param_names_);

        leave_scope ();
      }

    private:
      TemplateParamMemberDecl template_param_member_;
      Traversal::Names template_param_names_;

      TemplateParamBaseDecl template_param_base_;
      Traversal::Inherits template_param_inherits_;
    };


    //
    //
    struct TemplateParamUseList: Traversal::Complex,
                                 protected virtual Context
    {
      TemplateParamUseList (Context& c)
          : Context (c),
            template_param_member_ (c),
            template_param_base_ (c)
      {
        template_param_names_ >> template_param_member_;
        template_param_inherits_ >> template_param_base_;
      }

      virtual Void
      traverse (Type& c)
      {
        os << id ("_xsd_", c.name ());

        enter_class_scope (id (c.name ()), c);

        inherits (c, template_param_inherits_);
        names (c, template_param_names_);

        leave_scope ();
      }

    private:
      TemplateParamMemberUse template_param_member_;
      Traversal::Names template_param_names_;

      TemplateParamBaseUse template_param_base_;
      Traversal::Inherits template_param_inherits_;
    };


    //
    //
    struct ParserParamDecl : Traversal::Complex,
                             Traversal::Member,
                             protected virtual Context
    {
      ParserParamDecl (Context& c, Boolean name_arg)
          : Context (c), first_ (true), name_arg_ (name_arg)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> *this;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        enter_class_scope (id (c.name ()), c);
        Traversal::Complex::traverse (c);
        leave_scope ();
      }

      virtual Void
      traverse (SemanticGraph::Member& m)
      {
        String name (m.name ());

        if (!first_)
          os << ", ";
        else
          first_ = false;

        os << "::xsd::cxx_cli::parser::Traverser< " <<
          id ("_xsd_", name) << " >^";

        if (name_arg_)
          os << " " << id (name);
      }

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;

      Boolean first_;
      Boolean name_arg_;
    };


    //
    //
    struct Base : Traversal::Complex,
                  protected virtual Context
    {
      Base (Context& c)
          : Context (c), base_ (c), member_ (c)
      {
        *this >> inherits_ >> base_;
        *this >> names_ >> member_;
      }

      virtual Void
      traverse (Type& c)
      {
        String type (fq_name (c));

        os << type << "< " << id ("_xsd_", c.name ());

        enter_class_scope (id (c.name ()), c);

        Traversal::Complex::traverse (c);

        leave_scope ();

        os << " >";
      }

    private:
      TemplateParamBaseUse base_;
      Traversal::Inherits inherits_;


      TemplateParamMemberUse member_;
      Traversal::Names names_;
    };
  }
}

#endif  // CXX_CLI_PARSER_ELEMENTS_HXX
