// file      : xsd/cxx-cli/parser/validator.cxx-cli
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx-cli/parser/validator.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cxx-cli/parser/elements.hxx>

#include <iostream>

using std::wcerr;

namespace CXX_CLI
{
  namespace Parser
  {
    namespace
    {
      class ValidationContext: public Context
      {
      public:
        ValidationContext (CLI::Options const& options, Boolean& valid_)
            : Context (std::wcerr, options),
              valid (valid_)
        {
        }

      public:
        String
        xpath (SemanticGraph::Nameable& n)
        {
          if (dynamic_cast<SemanticGraph::Namespace*> (&n) != 0)
            return L"<namespace-level>"; // There is a bug if you see this.

          if (n.named ())
          {
            SemanticGraph::Scope& scope (n.scope ());

            if (dynamic_cast<SemanticGraph::Namespace*> (&scope) != 0)
              return n.name ();

            return xpath (scope) + L"/" + n.name ();
          }
          else
          {
            return L"(anonymous type for " +
              n.context ().get<String> ("instance-name") + L")";
          }
        }

      protected:
        ValidationContext (ValidationContext& c)
            :  Context (c),
               valid (c.valid)
        {
        }

      protected:
        Boolean& valid;
      };


      // Go into implied/included/imported schemas while making sure
      // we don't recurse too far.
      //
      struct Contains: Traversal::Contains
      {
        virtual Void
        traverse (Type& c)
        {
          SemanticGraph::Node& n (c.element ());

          if (!n.context ().count ("seen"))
          {
            n.context ().set ("seen", true);
            Traversal::Contains::traverse (c);
            n.context ().remove ("seen");
          }
        }
      };

      //
      //
      struct Any : Traversal::Any, protected virtual ValidationContext
      {
        Any (ValidationContext& c)
            : ValidationContext (c)
        {
        }

        struct Element: Traversal::Element,
                        protected virtual ValidationContext
        {
          Element (ValidationContext& c, SemanticGraph::Any& any)
              : ValidationContext (c),
                any_ (any),
                ns_ (xs_ns_name (any))
          {
          }

          virtual Void
          traverse (SemanticGraph::Element& e)
          {
            using SemanticGraph::Any;

            Boolean q (e.qualified ());
            String ns (q ? e.namespace_ ().name () : "");

            for (Any::NamespaceIterator i (any_.namespace_begin ());
                 i != any_.namespace_end (); ++i)
            {
              Boolean failed (false);

              if (*i == L"##any")
              {
                failed = true;
              }
              else if (*i == L"##other")
              {
                if (ns_)
                {
                  // Note that here I assume that ##other does not
                  // include names without target namespace. This
                  // is not what the spec says but that seems to be
                  // the consensus.
                  //
                  failed = q && ns != ns_;
                }
                else
                {
                  // No target namespace.
                  //
                  failed = q && ns != L"";
                }
              }
              else if (*i == L"##local")
              {
                failed = !q || ns == L"";
              }
              else if (*i == L"##targetNamespace")
              {
                failed = (q && ns_ == ns) || (!q && ns_ == L"");
              }
              else
              {
                failed = q && *i == ns;
              }

              if (failed)
              {
                Any& a (any_);

                os << a.file () << ":" << a.line () << ":" << a.column ()
                   << ": error: namespace '" << *i << "' allows for element "
                   << "'" << e.name () << "'" << endl;

                os << e.file () << ":" << e.line () << ":" << e.column ()
                   << ": error: element '" << e.name () << "' is defined "
                   << "here" << endl;

                os << a.file () << ":" << a.line () << ":" << a.column ()
                   << ": error: such wildcards are not supported" << endl;

                valid = false;
              }
            }
          }

        private:
          SemanticGraph::Any& any_;
          String ns_;
        };

        struct Complex: Traversal::Complex
        {
          Complex ()
              : up_ (true), down_ (true)
          {
          }

          virtual Void
          post (Type& c)
          {
            // Go down the inheritance hierarchy.
            //
            if (down_)
            {
              Boolean up = up_;
              up_ = false;

              for (Type::InheritsIterator i (c.inherits_begin ());
                   i != c.inherits_end (); ++i)
              {
                dispatch (i->inheritee ());
              }

              up_ = up;
            }

            // Go up the inheritance hierarchy.
            //
            if (up_)
            {
              Boolean down = down_;
              down_ = false;

              for (Type::BegetsIterator i (c.begets_begin ());
                   i != c.begets_end (); ++i)
              {
                dispatch (i->inheritor ());
              }

              down_ = down;
            }
          }

        private:
          Boolean up_, down_;
        };

        virtual Void
        traverse (SemanticGraph::Any& a)
        {
          SemanticGraph::Complex& type (
            dynamic_cast<SemanticGraph::Complex&> (a.scope ()));

          Complex complex;
          Traversal::Names names;
          Element element (*this, a);

          complex >> names >> element;

          complex.dispatch (type);
        }
      };

      //
      //
      struct Traverser : Traversal::Schema,
                         Traversal::Complex,
                         Traversal::Type,
                         Any,
                         Traversal::Element,
                         Traversal::Attribute,
                         protected virtual ValidationContext
      {
        Traverser (ValidationContext& c)
            : Any (c),
              ValidationContext (c),
              anonymous_ (c, *this)
        {
          edge_traverser (contains_);
          edge_traverser (schema_names_);

          contains_.node_traverser (*this);
          schema_names_.node_traverser (ns_);

          ns_.edge_traverser (names_);
          names_.node_traverser (*this);
          names_.node_traverser (anonymous_);

          edge_traverser (names_);
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          traverse (static_cast<SemanticGraph::Type&> (c));

          //@@ Maybe use traverser?
          //
          SemanticGraph::Complex::InheritsIterator it (c.inherits_begin ());

          if (it != c.inherits_end ())
          {
            SemanticGraph::Type& t (it->inheritee ());

            if (t.named () &&
                types_.find (
                  t.scope ().name () + L"#" + t.name ()) == types_.end ())
            {
              valid = false;

              wcerr << c.file () << ":" << c.line () << ":" << c.column ()
                    << ": error: type '" << xpath (c) << "' inherits from "
                    << "yet undefined type '" << xpath (t) << "'" << endl;

              wcerr << t.file () << ":" << t.line () << ":" << t.column ()
                    << ": info: '" << xpath (t) << "' is defined here"
                    << endl;

              wcerr << c.file () << ":" << c.line () << ":" << c.column ()
                    << ": info: inheritance from a yet-undefined type is not "
                    << "supported" << endl;

              wcerr << c.file () << ":" << c.line () << ":" << c.column ()
                    << ": info: re-arrange your schema and try again" << endl;
            }
          }

          Traversal::Complex::traverse (c);
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          if (t.named ())
          {
            types_.insert (t.scope ().name () + L"#" + t.name ());
          }
        }

        virtual Void
        traverse (SemanticGraph::Element& e)
        {
          /*
          SemanticGraph::Type& t (e.type ());

          // Check for forward inheritance.
          //
          if (dynamic_cast<SemanticGraph::AnyType*> (&t) != 0 &&
              dynamic_cast<SemanticGraph::Fundamental::Type*> (&t) == 0)
          {
            valid = false;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": error: element '" << xpath (e) << "' is of type "
                  << "'anyType'" << endl;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": info: 'anyType' is not supported" << endl;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": info: did you forget to specify 'type' attribute?"
                  << endl;
          }
          */
        }


        virtual Void
        traverse (SemanticGraph::Attribute& a)
        {
          /*
          SemanticGraph::Type& t (a.type ());

          if (dynamic_cast<SemanticGraph::AnySimpleType*> (&t) != 0 &&
              dynamic_cast<SemanticGraph::Fundamental::Type*> (&t) == 0)
          {
            valid = false;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": error: attribute '" << xpath (a) << "' is of type "
                  << "'anySimpleType'" << endl;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": info: 'anySimpleType' is not supported" << endl;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": info: did you forget to specify 'type' attribute?"
                  << endl;
          }
          */
        }

      private:
        Containers::Set<String> types_;

        Contains contains_;

        Traversal::Names schema_names_;
        Traversal::Namespace ns_;

        Traversal::Names names_;

        // Anonymous feedback via belongs edge.
        //
        struct Anonymous : AnonymousBase,
                           protected virtual ValidationContext
        {
          Anonymous (ValidationContext& c, Traversal::NodeDispatcherBase& d)
              : ValidationContext (c),
                AnonymousBase (d)
          {
          }

          virtual void
          member_pre (SemanticGraph::Member& m)
          {
            m.type ().context ().set ("instance-name", xpath (m));
          }

          virtual void
          member_post (SemanticGraph::Member& m)
          {
            m.type ().context ().remove ("instance-name");
          }

        } anonymous_;
      };


      struct AnonymousType : Traversal::Schema,
                             Traversal::Complex,
                             Traversal::Element,
                             Traversal::Attribute,
                             protected virtual ValidationContext
      {
        AnonymousType (ValidationContext& c)
            : ValidationContext (c)
        {
          *this >> contains_ >> *this;


          *this >> schema_names_ >> ns_ >> names_ >> *this;
          *this >> names_;
        }

        virtual Void
        traverse (SemanticGraph::Element& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (!t.named ()
              && dynamic_cast<SemanticGraph::Fundamental::IdRef*> (&t) == 0
              && dynamic_cast<SemanticGraph::Fundamental::IdRefs*> (&t) == 0)
          {
            valid = false;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": error: element '" << xpath (e) << "' "
                  << "is of anonymous type" << endl;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": info: "
                  << "anonymous types are not supported in this mapping"
                  << endl;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": info: "
                  << "use --morph-anonymous to automatically name "
                  << "anonymous types"
                  << endl;
          }
          else
          {
            Traversal::Element::traverse (e);
          }
        }

        virtual Void
        traverse (SemanticGraph::Attribute& a)
        {
          SemanticGraph::Type& t (a.type ());

          if (!t.named ()
              && dynamic_cast<SemanticGraph::Fundamental::IdRef*> (&t) == 0
              && dynamic_cast<SemanticGraph::Fundamental::IdRefs*> (&t) == 0)
          {
            valid = false;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": error: attribute '" << xpath (a) << "' "
                  << "is of anonymous type" << endl;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": info: "
                  << "anonymous types are not supported in this mapping"
                  << endl;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": info: "
                  << "use --morph-anonymous to automatically name "
                  << "anonymous types"
                  << endl;
          }
          else
          {
            Traversal::Attribute::traverse (a);
          }
        }

      private:
        Containers::Set<String> types_;

        Contains contains_;

        Traversal::Names schema_names_;
        Traversal::Namespace ns_;

        Traversal::Names names_;
      };
    }

    Validator::
    Validator ()
    {
      // Dummy ctor, help with long symbols on HP-UX.
    }

    Boolean Validator::
    validate (CLI::Options const& options,
              SemanticGraph::Schema& schema,
              SemanticGraph::Path const& tu)
    {
      Boolean valid (true);

      ValidationContext ctx (options, valid);

      // Test for anonymout types.
      //
      {
        AnonymousType traverser (ctx);

        traverser.dispatch (schema);
      }


      // Test the rest.
      //
      if (valid)
      {
        Traverser traverser (ctx);

        traverser.dispatch (schema);
      }

      return valid;
    }
  }
}

