// file      : xsd/cxx/tree/serialization-header.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/serialization-header.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      struct List: Traversal::List, protected virtual Context
      {
        List (Context& c,
              String const& name_ = "",
              String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& l)
        {
          Boolean named (l.named ());

          if (named)
          {
            name = l.name ();

            enter_scope (id (name));
            type_name = scope;
            leave_scope ();
          }

          // operator<< (xml::element)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (" << xml_element_type << "&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;

          // operator<< (xml::attribute)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (" << xml_attribute_type << "&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;

          // operator<< (list_stream)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xsd::cxx::xml::dom::list_stream< " << char_type
             << " >&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;

        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      struct Union: Traversal::Union, protected virtual Context
      {
        Union (Context& c,
               String const& name_ = "",
               String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& u)
        {
          Boolean named (u.named ());

          if (named)
          {
            name = u.name ();

            enter_scope (id (name));
            type_name = scope;
            leave_scope ();
          }

          // operator<< (xml::element)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (" << xml_element_type << "&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;

          // operator<< (xml::attribute)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (" << xml_attribute_type << "&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;

          // operator<< (list_stream)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xsd::cxx::xml::dom::list_stream< " << char_type
             << " >&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      struct Enumeration: Traversal::Enumeration, protected virtual Context
      {
        Enumeration (Context& c,
                     String const& name_ = "",
                     String const& type_name_ = "")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& e)
        {
          Boolean named (e.named ());

          if (named)
          {
            name = e.name ();

            enter_class_scope (id (name), e);
            type_name = scope;
            leave_scope ();
          }

          // operator<< (xml::element)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (" << xml_element_type << "&," << endl
             << (named ? "class " : "") << type_name << ");"
             << endl;

          // operator<< (xml::attribute)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (" << xml_attribute_type << "&," << endl
             << (named ? "class " : "") << type_name << ");"
             << endl;

          // operator<< (list_stream)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (::xsd::cxx::xml::dom::list_stream< " << char_type
             << " >&," << endl
             << (named ? "class " : "") << type_name << ");"
             << endl;
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      //
      //
      template <typename Complex>
      struct Member : Traversal::Member, protected virtual Context
      {
        Member (Context& c, String const& scope_)
            : Context (c),
              scope (scope_)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          SemanticGraph::Type& t (m.type ());

          if (!t.named () && !t.context ().count ("seen"))
          {
            // I need to build scope information with the same names
            // as in the header so that I get member escaping right.
            // Actual names will be from the public interface.
            //

            String name (m.name ());
            String escaped_name (id (name));
            String escaped_xsd_name (id ("_xsd_", name));
            String escaped_xsd_name_ (id ("_xsd_", name, "_"));

            String type_name (scope
                              + L"::" + escaped_name
                              + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, escaped_name, type_name);
            Union union_ (*this, escaped_name, type_name);
            Complex complex (*this, escaped_name, type_name);
            Enumeration enumeration (*this, escaped_name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (escaped_xsd_name);
            enter_scope (escaped_xsd_name_);

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }

      private:
        String scope;
      };


      struct Complex : Traversal::Complex, protected virtual Context
      {
        Complex (Context& c,
                 String const& name_ = "",
                 String const& type_name_ = L"")
            : Context (c),
              name (name_),
              type_name (type_name_),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }

        virtual Void
        traverse (Type& c)
        {
          Boolean named (c.named ());

          if (named)
          {
            name = c.name ();

            enter_class_scope (id (name), c);
            type_name = scope;
            leave_scope ();
          }

          // Go after anonymous types first.
          //
          {
            enter_class_scope (id (name), c);

            Traversal::Names names_member;
            Member<Complex> member (*this, type_name);

            names_member >> member;

            names (c, names_member);

            leave_scope ();
          }


          // operator<< (xml::element)
          //
          os << esymbol
             << "void" << endl
             << "operator<< (" << xml_element_type << "&," << endl
             << (named ? "class " : "") << type_name << " const&);"
             << endl;


          //@@ Need to factor this out into serialization-elements
          //
          Boolean simple (true);

          struct Traverser : Traversal::Complex
          {
            Traverser (Boolean& v)
                : v_ (v)
            {
              edge_traverser (inherits_);
              inherits_.node_traverser (*this);
            }

            virtual Void
            traverse (SemanticGraph::Complex& c)
            {
              if (has<Traversal::Element> (c) ||
                  has<Traversal::Attribute> (c))
                v_ = false;
              else
                inherits (c, *this);
            }

          private:
            Traversal::Inherits inherits_;
            Boolean& v_;
          } t (simple);

          t.dispatch (c);

          if (simple)
          {
            // operator<< (xml::attribute)
            //
            os << esymbol
               << "void" << endl
               << "operator<< (" << xml_attribute_type << "&," << endl
               << (named ? "class " : "") << type_name << " const&);"
               << endl;

            // operator<< (list_stream)
            //
            os << esymbol
               << "void" << endl
               << "operator<< (::xsd::cxx::xml::dom::list_stream< " << char_type
               << " >&," << endl
               << (named ? "class " : "") << type_name << " const&);"
               << endl;
          }
        }

      private:
        String name;
        String type_name;
        String esymbol;
      };


      // Goes after anonymous types via global element. Note that we don't
      // need to go after attributes because they can only be referenced
      // and in that case everything will be generated at the point of
      // reference.
      //
      struct GlobalMember : Traversal::Element, protected virtual Context
      {
        GlobalMember (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          SemanticGraph::Type& t (m.type ());

          if (!t.named ())
          {
            String name (m.name ());
            String escaped_name (id (name));
            String escaped_xsd_name_ (id ("_xsd_", name, "_"));

            String type_name (escaped_name + L"::" + unclash (name, "type"));

            Traversal::Belongs belongs_;

            List list (*this, escaped_name, type_name);
            Union union_ (*this, escaped_name, type_name);
            Complex complex (*this, escaped_name, type_name);
            Enumeration enumeration (*this, escaped_name, type_name);

            belongs_ >> list;
            belongs_ >> union_;
            belongs_ >> complex;
            belongs_ >> enumeration;

            enter_scope (escaped_name);
            enter_scope (escaped_xsd_name_);

            t.context ().set ("seen", true);

            belongs (m, belongs_);

            t.context ().remove ("seen");

            leave_scope ();
            leave_scope ();
          }
        }
      };


      struct Serializer: Traversal::Element,
                         protected virtual Context
      {
        Serializer (Context& c)
            : Context (c),
              esymbol (c.esymbol.empty () ? c.esymbol : c.esymbol + L"\n")
        {
        }


        virtual Void
        traverse (Type& e)
        {
          String name (e.name ());

          os << "// Serialize to an existing DOM instance." << endl
             << "//" << endl
             << esymbol
             << "void" << endl
             << id (name) << " (::xercesc::DOMDocument&," << endl
             << type_name (e) << " const&," << endl
             << "::xsd::cxx::tree::flags = 0);"
             << endl
             << endl;

          os << "// Serialize to a new DOM instance." << endl
             << "//" << endl
             << esymbol
             << "::xsd::cxx::xml::dom::auto_ptr< ::xercesc::DOMDocument >" << endl
             << id (name) << " (" << type_name (e) << " const&, " << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const&," << endl
             << "::xsd::cxx::tree::flags = 0);"
             << endl
             << endl;

          os << "// Serialize to XMLFormatTarget." << endl
             << "//" << endl
             << esymbol
             << "void" << endl
             << id (name) << " (::xercesc::XMLFormatTarget&," << endl
             << type_name (e) << " const&, " << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const&," << endl
             << "::std::basic_string< " << char_type << " > const& = " << L <<
            "\"UTF-8\"," << endl
             << "::xsd::cxx::tree::flags = 0);"
             << endl
             << endl;

          os << esymbol
             << "void" << endl
             << id (name) << " (::xercesc::XMLFormatTarget&," << endl
             << type_name (e) << " const&, " << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const&," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >&," << endl
             << "::std::basic_string< " << char_type << " > const& = " << L <<
            "\"UTF-8\"," << endl
             << "::xsd::cxx::tree::flags = 0);"
             << endl;


          os << esymbol
             << "void" << endl
             << id (name) << " (::xercesc::XMLFormatTarget&," << endl
             << type_name (e) << " const&, " << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const&," << endl
             << "::xercesc::DOMErrorHandler&," << endl
             << "::std::basic_string< " << char_type << " > const& = " << L <<
            "\"UTF-8\"," << endl
             << "::xsd::cxx::tree::flags = 0);"
             << endl
             << endl;

          os << "// Serialize to std::ostream." << endl
             << "//" << endl
             << esymbol
             << "void" << endl
             << id (name) << " (::std::ostream&," << endl
             << type_name (e) << " const&, " << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const&," << endl
             << "::std::basic_string< " << char_type << " > const& = " << L <<
            "\"UTF-8\"," << endl
             << "::xsd::cxx::tree::flags = 0);"
             << endl
             << endl;

          os << esymbol
             << "void" << endl
             << id (name) << " (::std::ostream&," << endl
             << type_name (e) << " const&, " << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const&," << endl
             << "::xsd::cxx::xml::error_handler< " <<
            char_type << " >&," << endl
             << "::std::basic_string< " << char_type << " > const& = " << L <<
            "\"UTF-8\"," << endl
             << "::xsd::cxx::tree::flags = 0);"
             << endl;

          os << esymbol
             << "void" << endl
             << id (name) << " (::std::ostream&," << endl
             << type_name (e) << " const&, " << endl
             << "::xsd::cxx::xml::dom::namespace_infomap< " << char_type <<
            "> const&," << endl
             << "::xercesc::DOMErrorHandler&," << endl
             << "::std::basic_string< " << char_type << " > const& = " << L <<
            "\"UTF-8\"," << endl
             << "::xsd::cxx::tree::flags = 0);"
             << endl
             << endl;

        }

      private:
        String
        type_name (Type& e)
        {
          SemanticGraph::Type& t (e.type ());

          if (t.named ())
          {
            std::wostringstream o;

            MemberTypeName type (*this, o);
            type.dispatch (t);

            return o.str ();
          }
          else
          {
            String name (id (e.name ()));

            return name + L"::" + unclash (name, "type");
          }
        }

      private:
        String esymbol;
      };

    }

    Void
    generate_serialization_header (Context& ctx, SemanticGraph::Schema& tu)
    {
      String c (ctx.char_type);

      ctx.os << "#include <iosfwd> // std::ostream&" << endl
             << endl
             << "#include <xsd/cxx/tree/serialization.hxx>" << endl
             << endl;


      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names names_ns, names;

      Namespace ns (ctx);

      List list (ctx);
      Union union_ (ctx);
      Complex complex (ctx);
      Enumeration enumeration (ctx);
      GlobalMember member (ctx);
      Serializer serializer (ctx);


      schema >> sources >> schema;
      schema >> names_ns >> ns >> names;

      names >> list;
      names >> union_;
      names >> complex;
      names >> enumeration;
      names >> member;
      names >> serializer;

      schema.dispatch (tu);
    }
  }
}
